/**
 *  Naken Chat
 *  Author: Michael Kohn
 *   Email: mike@mikekohn.net
 *     Web: http://www.mikekohn.net/
 * License: GPL
 *
 * Copyright 1998-2013 by Michael Kohn
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#ifndef WINDOWS
#include <pthread.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#include <netdb.h>
#endif
#include <errno.h>
#include <time.h>
#include <string.h>
#ifdef WINDOWS
#include <windows.h>
#include <process.h>
#include <winsock.h>
#endif

#include "chat_thread.h"
#include "general.h"
#include "info.h"
#include "language.h"
#include "message.h"
#include "nakenchat.h"
#include "passwd.h"
#include "read_config.h"
#include "version.h"

struct _server_info *server_info;

struct _user **users;
struct _user nulluser;

#ifdef PLUGINS
struct naken_plugins_t *naken_plugins=0;
struct naken_context_t naken_context;
#endif

#define SSL_OFF 0
#define SSL_ON 1

static int init_user(int socketid, struct sockaddr_in *cli_addr, socklen_t clilen, int use_ssl)
{
int k,r,t,ID;
char temp_string[200];
struct hostent *my_hostent=NULL;
unsigned char *address=0;
char *fullmessage={ SORRY_FULL };
char *toomanyconnections={ TOO_MANY_CONNECTIONS };
char *allowptr;
char *banptr;
#ifdef ENABLE_SSL
SSL *ssl=NULL;
#endif

  if (server_info->flags.host_lookups)
  {
    my_hostent=gethostbyaddr((char *)&(cli_addr->sin_addr.s_addr),4,AF_INET);
  }

#ifdef ENABLE_SSL
  if (use_ssl==SSL_ON)
  {
    ssl=SSL_new(server_info->ssl_ctx);
    SSL_set_fd(ssl, socketid);

    if (SSL_accept(ssl)<=0)
    {
#ifdef DEBUG
      if (server_info->debug==1)
      {
        ERR_print_errors_fp(stderr);
      }
#endif

      SSL_free(ssl);
      socketdie(socketid);
      return -1;
    }
  }
#endif

  for (r=0; r<server_info->max_conn; r++)
  { 
    if (users[r]->inuse==0) 
    { break; } 
  }

  if (r>=server_info->max_conn)
  {
#ifdef DEBUG
    if (server_info->debug==1) printf("Chat server full\n");
#endif
#ifdef ENABLE_SSL
    if (use_ssl==SSL_ON)
    {
      SSL_write(ssl, fullmessage, sizeof(fullmessage)-1); 
      SSL_free(ssl);
    }
      else
#endif
    {
      send(socketid, fullmessage, sizeof(fullmessage)-1, 0); 
    }
    socketdie(socketid);
    return -1;
  }

  // If this user is pointing to the empty user line, then we must malloc.
  if (users[r]==&nulluser)
  { 
#ifdef DEBUG
    if (server_info->debug==1) printf("Allocing new user %d\n",r);
#endif
    users[r]=(struct _user *)malloc(sizeof(struct _user)); 
    memset(users[r],0,sizeof(struct _user));
    users[r]->gags=malloc(server_info->gag_size); 
  }

  users[r]->uflags=0;
#ifdef ENABLE_SSL
  users[r]->flags.ssl_connected=(use_ssl==0)?0:1;
  users[r]->ssl=ssl;
#endif

#ifdef DEBUG
  if (server_info->debug==1)
  {
    printf("searched for open User: found %d\n",r);
    fflush(stdout);
  }
#endif

  if (my_hostent!=NULL && server_info->flags.host_lookups)
  {
    safestrncpy(users[r]->location,my_hostent->h_name,61);
  }
    else
  { 
    address=(unsigned char*)&cli_addr->sin_addr.s_addr;
    sprintf(users[r]->location,"%d.%d.%d.%d",address[0],address[1],address[2],address[3]);
  }

#ifdef DEBUG
  if (server_info->debug==1)
  {
    printf("Connected from location: %s\n",users[r]->location);
    fflush(stdout);
  }
#endif

  if (server_info->flags.no_dup_ips)
  {
    for (t=0; t<server_info->max_conn; t++)
    { 
      if (t==r) continue;

      if (users[t]->inuse!=0) 
      {
        if (strcmp(users[t]->location,users[r]->location)==0)
        {
#ifdef DEBUG
          if (server_info->debug==1) printf("Too many connections from this IP\n");
#endif

#ifdef ENABLE_SSL
          if (use_ssl==SSL_ON)
          {
            SSL_write(ssl, toomanyconnections, sizeof(toomanyconnections)-1); 
            SSL_free(ssl);
          }
            else
#endif
          {
            send(socketid, toomanyconnections, sizeof(toomanyconnections)-1,0); 
          }
          socketdie(socketid);
          return -1;
        }
      } 
    }
  }

  if (server_info->login_speed!=0)
  {
    k=time(NULL);
    for (t=0; t<server_info->max_conn; t++)
    { 
      if (t==r) continue;

      if (users[t]->inuse!=0) 
      {
        if ((k-users[t]->logontime)<server_info->login_speed && strcmp(users[t]->location,users[r]->location)==0)
        {
#ifdef DEBUG
          if (server_info->debug==1) printf("Fast login - user kicked\n");
#endif
#ifdef ENABLE_SSL
          if (use_ssl==SSL_ON)
          {
            SSL_write(ssl, toomanyconnections, sizeof(toomanyconnections)-1); 
            SSL_free(ssl);
          }
            else
#endif
          {
            send(socketid, toomanyconnections, sizeof(toomanyconnections)-1,0); 
          }
          socketdie(socketid);
          return -1;
        }
      } 
    }
  }

  banptr=server_info->bans;
  for (t=0; t<server_info->ban_count; t++)
  { 
    if (strcasestr(users[r]->location,banptr)!=NULL)
    {
#ifdef ENABLE_SSL
      if (use_ssl==SSL_ON)
      {
        SSL_write(ssl, BANNED, sizeof(BANNED)-1); 
        SSL_free(ssl);
      }
        else
#endif
      {
        send(socketid, BANNED, sizeof(BANNED)-1, 0);
      }
#ifdef DEBUG
      if (server_info->debug==1)
      { 
        printf("This site was banned(%d):\n",socketdie(socketid)); 
        fflush(stdout); 
      }
#endif

      socketdie(socketid);
      return -1;
    }

    banptr=banptr+strlen(banptr)+1;
  }

  if (server_info->allow_count!=0 && t==server_info->ban_count)
  {
    allowptr=server_info->allows;
    for (t=0; t<server_info->allow_count; t++)
    {
      if (strcasestr(users[r]->location,allowptr)!=NULL) break;
      allowptr=allowptr+strlen(allowptr)+1;
    }

    if (t==server_info->allow_count) 
    {
#ifdef ENABLE_SSL
      if (use_ssl==SSL_ON)
      {
        SSL_write(ssl, REFUSED, sizeof(REFUSED)-1); 
        SSL_free(ssl);
      }
        else
#endif
      {
        send(socketid, REFUSED, sizeof(REFUSED)-1, 0);
      }
#ifdef DEBUG
      if (server_info->debug==1)
      { 
        printf("This site not allowed(%d):\n",socketdie(socketid)); 
        fflush(stdout); 
      }
#endif

      socketdie(socketid);
      return -1;
    }
  }

/*
  k=0;

  if (server_info->flags.no_dup_ips)
  {
    for (k=0; k<server_info->max_conn; k++)
    {
      if (users[k]->inuse!=1) continue;
      if (strcmp(users[r]->location,users[k]->location)==0)
      { k=-1; break; }
    }
  }
*/

  users[r]->ID=r;
  users[r]->socketid=socketid;
  ID=r;

  users[ID]->spam=0;
  users[ID]->channel=0;
  users[ID]->user_level=0;
  //users[ID]->uflags=USER_ECHOS;
  users[ID]->flags.echos=1;
  users[ID]->stamptime=0;
  users[ID]->lastp=-1;
  users[ID]->ptr=0;
  users[ID]->blen=0;
  users[ID]->bptr=0;
  users[ID]->emote_char=EMOTE_CHAR;
  users[ID]->logontime=time(NULL);
  users[ID]->idletime=time(NULL);
#ifdef REALIDLETIME
  users[ID]->real_idletime=time(NULL);
#endif
  safestrncpy(users[ID]->username,users[ID]->location,MAX_USER_NAME_LENGTH);
  memset(users[ID]->gags,0,server_info->gag_size);

#ifndef WINDOWS
  fcntl(users[ID]->socketid, F_SETFL, O_NONBLOCK);
#else
  u_long arg=1L;
  ioctlsocket(users[ID]->socketid, FIONBIO, (u_long FAR *)&arg);
#endif

/* hmmm 

  if (setsockopt(sockfd,SOL_SOCKET,SO_RCVLOWAT,&sopt,sizeof(sopt)))
  { printf("socket options SO_RCVLOWAT can't be set.\n"); }
  if (setsockopt(sockfd,SOL_SOCKET,SO_SNDLOWAT,&sopt,sizeof(sopt)))
  { printf("socket options SO_SNDLOWAT can't be set.\n"); }
*/
 
  t=0;
  while((users[ID]->username[t]!=0 && users[ID]->username[t]!='.') && t<16) t++;
  users[ID]->username[t]=0; 

  if (server_info->flags.no_dup_names)
  {
    if (checkname_dup(users[ID]->username,ID)!=-1)
    { strcpy(users[ID]->username,"unnamed"); }
  }

#ifdef PASSWD
  if (server_info->passwd[0]!=0)
  {
    if (passwd_check(users[ID]->username,-1)==1)
    { strcpy(users[ID]->username,"unnamed"); }
  }
#endif

  for (r=0; r<server_info->gag_size; r++)
  {
    users[ID]->gags[r]=0;
  }

#ifdef ENABLE_SSL
  if (use_ssl==SSL_ON)
  {
    SSL_write(users[ID]->ssl, server_info->welcome, server_info->welcome_len); 
  }
    else
#endif
  {
    send(users[ID]->socketid, server_info->welcome, server_info->welcome_len,0);
  }

//printf("users[%d]->flags.ssl_connected=%d\n", ID, users[ID]->flags.ssl_connected);
//printf(":(\n");
  if (server_info->flags.force_names)
  {
/*
    sprintf(temp_string,NEED_TO_LOGIN,ID,users[ID]->location);
    message(ID,temp_string,0);
*/

    users[ID]->inuse=100;
  }
    else
  {
    users[ID]->inuse=1;

    sprintf(temp_string, YOU_LOGIN, ID, users[ID]->location);
    message(ID, temp_string, 0);

    if (server_info->flags.show_ips)
    {
#ifdef ENABLE_SSL
      if (use_ssl==1)
      {
        sprintf(temp_string, SOMEONE_LOGIN1_SSL, ID, users[ID]->location);
      }
        else
#endif
      {
        sprintf(temp_string, SOMEONE_LOGIN1, ID, users[ID]->location);
      }
    }
      else
    {
#ifdef ENABLE_SSL
      if (use_ssl==1)
      {
        sprintf(temp_string, SOMEONE_LOGIN2, ID);
      }
        else
#endif
      {
        sprintf(temp_string, SOMEONE_LOGIN2, ID);
      }
    }

    sendallbut(ID,temp_string);
    update_list(ID,1);
  }

#ifdef WHOFILE
  if (server_info->whofile[0]!=0) write_whofile();
#endif

  return 0;
}

void kill_user(int ID)
{
int r;
char spa[80];

  if (users[ID]->inuse==5) return;
  if (users[ID]->inuse==1) users[ID]->inuse=5;

#ifdef DEBUG
  if (server_info->debug==1)
  {
    printf("Getting ready to close socket.\n");
    if (users[ID]->inuse!=4)
      printf("Socket Close Return value: %d:\n",socketdie(users[ID]->socketid));
  }
    else
#endif
  {
    if (users[ID]->inuse!=4) socketdie(users[ID]->socketid);
  }

  users[ID]->socketid=0;

  if (users[ID]->channel!=0) channel(ID,"0");

  for (r=0; r<MAXCHAN; r++) 
  { if (server_info->channel_op[r]==ID) server_info->channel_op[r]=-1; }
/*
  curr=users[ID]->channel;
  if (curr!=0)
  {
    if (server_info->channel_op[curr]==ID) server_info->channel_op[curr]=-1;
    for (r=0; r<server_info->max_conn; r++)
    {
      if (users[r]->inuse==1)
      {
        if (users[r]->channel==curr && r!=ID) r=server_info->max_conn+6;
      }
    }

    if (r<server_info->max_conn)
    { channels[curr][0]=0; server_info->channel_op[curr]=-1; server_info->channel_flags[curr]=0; }
  }
*/

  for (r=0; r<server_info->max_conn; r++)
  {
    if (users[r]->inuse==1)
    {
      if (IS_GAGGED(r,ID))
      {
        TOGGLE_GAG(r,ID);
      }
    }
  }
  for (r=0; r<server_info->max_conn; r++) if (users[r]->inuse==1 && users[r]->lastp==ID) users[r]->lastp=-1;

#ifdef DEBUG
  if (server_info->debug==1) printf("Connection dead: %d\n",ID);
  fflush(stdout);
#endif

  if (users[ID]->inuse!=100)
  {
    users[ID]->inuse=0;
    sprintf(spa,JUST_LEFT,ID,users[ID]->username);

    if (!server_info->flags.channel_disallow)
    { sendallbut(ID,spa); }
      else
    { sendchanbut(ID,users[ID]->channel,spa); }

    update_list(ID,0);
  }
    else
  { users[ID]->inuse=0; }

  users[ID]->idletime=time(NULL);
#ifdef REALIDLETIME
  users[ID]->real_idletime=time(NULL);
#endif

#ifdef WHOFILE
  if (server_info->whofile[0]!=0) write_whofile();
#endif
}

#ifdef ENABLE_SSL
static int init_ssl()
{
  SSL_library_init();
  OpenSSL_add_all_algorithms();
  SSL_load_error_strings();
  //server_info->ssl_method=SSLv3_server_method();
  const SSL_METHOD *ssl_method=SSLv3_server_method();
  server_info->ssl_ctx=SSL_CTX_new(ssl_method);

  if (server_info->ssl_ctx==NULL)
  {
    ERR_print_errors_fp(stderr);
    exit(1);
  }

  if (SSL_CTX_use_certificate_file(server_info->ssl_ctx, server_info->ssl_cert, SSL_FILETYPE_PEM)<=0)
  {
    ERR_print_errors_fp(stderr);
    exit(1);
  }

  if (SSL_CTX_use_PrivateKey_file(server_info->ssl_ctx, server_info->ssl_cert, SSL_FILETYPE_PEM)<=0)
  {
    ERR_print_errors_fp(stderr);
    exit(1);
  }

  if (!SSL_CTX_check_private_key(server_info->ssl_ctx))
  {
    fprintf(stderr, "Private key does not match the public certificate\n");
    exit(1);
  }

  return 0;
}
#endif

static int open_server_port(int port)
{
//struct linger mylinger;
struct sockaddr_in serv_addr;
int sockfd;
int sopt=1;

  if ((sockfd=socket(AF_INET, SOCK_STREAM, 0))<0)
  {
    printf("Can't open socket.\n");
    exit(1);
  }

  /* bzero((char*)&serv_addr,sizeof(serv_addr)); */
  memset((char*)&serv_addr,0,sizeof(serv_addr));
  serv_addr.sin_family=AF_INET;
  serv_addr.sin_addr.s_addr=htonl(INADDR_ANY);
  serv_addr.sin_port=htons(port);

  //mylinger.l_onoff=0;
  //mylinger.l_linger=1;

#ifndef WINDOWS
  if (setsockopt(sockfd,SOL_SOCKET,SO_REUSEADDR,&sopt,sizeof(sopt)))
  { printf("socket options REUSERADDR can't be set.\n"); }
  //if (setsockopt(sockfd,SOL_SOCKET,SO_LINGER,&mylinger,sizeof(struct linger)))
  //{ printf("Socket options LINGER can't be set.\n"); }
#ifdef KEEPALIVE
  if (setsockopt(sockfd,SOL_SOCKET,SO_KEEPALIVE,&sopt,sizeof(sopt)))
  { printf("Socket options KEEPALIVE can't be set.\n"); }
#endif
#else
  if (setsockopt(sockfd,SOL_SOCKET,SO_REUSEADDR,(char *)&sopt,sizeof(sopt)))
  { printf("Socket options REUSERADDR can't be set.\n"); }
  //if (setsockopt(sockfd,SOL_SOCKET,SO_LINGER,(char *)&mylinger,sizeof(struct linger)))
  //{ printf("Socket options LINGER can't be set.\n"); }
#ifdef KEEPALIVE
  if (setsockopt(sockfd,SOL_SOCKET,SO_KEEPALIVE,(char *)&sopt,sizeof(sopt)))
  { printf("Socket options KEEPALIVE can't be set.\n"); }
#endif
#endif

  if(bind(sockfd,(struct sockaddr *)&serv_addr,sizeof(serv_addr))<0)
  {
    printf("Server can't bind.\n");
    exit(1);
  }

  listen(sockfd,5);

  return sockfd;
}

int main(int argc,char *argv[])
{
int sockfd=0;
#ifdef ENABLE_SSL
int sockfd_ssl=0;
#endif
int r;
int newsockfd;
struct sockaddr_in cli_addr;
int threads[MAX_USER_THREADS];
int socket_errors=0;
socklen_t clilen;
#ifndef WINDOWS
pthread_t pid;
#endif
#ifdef WINDOWS
WSADATA wsaData;
WORD wVersionRequested=MAKEWORD(1,1);
int Win32isStupid;

  Win32isStupid=WSAStartup(wVersionRequested, &wsaData);
  if (Win32isStupid)  /* Shouldn't this always be true? :) */
  {
    printf("Winsock can't start.\n");
    exit(1);
  }

/*
  printf("Max sockets: %d\n",wsaData.iMaxSockets);
  printf("Description: %s\n",wsaData.szDescription);
*/
#endif

  server_info=(struct _server_info *)malloc(sizeof(struct _server_info));
  memset(server_info, 0, sizeof(struct _server_info));
  server_info->max_conn=MAXCONN;
  server_info->min_conn=10;
  server_info->max_idle_time=0;
  server_info->register_level=-1;

  setup(argc,argv);

#ifdef DEBUG
  if (server_info->debug==1)
  {
    printf("min_conn: %d\n",server_info->min_conn);
    printf("max_conn: %d\n",server_info->max_conn);
  }
#endif

  if (server_info->port!=0)
  {
    sockfd=open_server_port(server_info->port);
  }

#ifdef ENABLE_SSL
  init_ssl();

  if (server_info->ssl_port!=0)
  {
    sockfd_ssl=open_server_port(server_info->ssl_port);
  }

  if (server_info->port==0 && server_info->ssl_port==0)
  {
    printf("Error: No server sockets opened.  Aborting.\n");
    exit(1);
  }
#endif

#ifndef WINDOWS
  setsignals();
#endif

#ifndef WINDOWS
  if (server_info->debug==0)
  {
    r=fork();

    if (r==-1)
    { perror("Error: Could not fork\n"); exit(3); }
      else
    if (r==0)
    {
      close(STDIN_FILENO);
      close(STDERR_FILENO);

      if (setsid()==-1)
      { exit(4); }
    }
      else
    { return 0; }
  }
#endif

/* YOU MAY NOT EDIT THE NEXT LINE OF CODE */
  printf("\n"VERSION"\n"COPYRIGHT"\n\n");
/* NAKEN CHAT IS NOT GPL */

  if (server_info->port!=0)
  {
    printf("Started on port: %d\n",server_info->port);
  }
#ifdef ENABLE_SSL
  if (server_info->ssl_port!=0)
  {
    printf("Started on port (with SSL): %d\n",server_info->ssl_port);
  }
#endif
  fflush(stdout);

  if (server_info->debug==0) close(STDOUT_FILENO);

  nulluser.inuse=0;
  nulluser.idletime=-1;
#ifdef REALIDLETIME
  nulluser.real_idletime=-1;
#endif
  users=malloc(sizeof(struct _user*)*server_info->max_conn);
#ifdef PLUGINS
  naken_context.users=users;
#endif

  for (r=0; r<server_info->max_conn; r++)
  {
    users[r]=&nulluser;
  }

#ifndef LOW_MEM_GAGS
  server_info->gag_size=server_info->max_conn;
#else
  server_info->gag_size=server_info->max_conn/8;
  if ((server_info->max_conn%8)!=0) server_info->gag_size++;
#endif

  for (r=0; r<server_info->min_conn; r++)
  {
    users[r]=(struct _user *)malloc(sizeof(struct _user));
    memset(users[r],0,sizeof(struct _user));
    // FIXME - fix this shit
    users[r]->gags=malloc(server_info->gag_size); 
    memset(users[r]->gags,0,server_info->gag_size);
    users[r]->inuse=0;
  }

#ifdef WHOFILE
#ifndef WINDOWS
  pthread_mutexattr_init(&server_info->pthread_mutexattr);
  pthread_mutex_init(&server_info->pthread_mutex, &server_info->pthread_mutexattr);
#else
  InitializeCriticalSection(&server_info->win_mutex);
#endif
#endif

  for (r=0; r<MAX_USER_THREADS; r++)
  {
    threads[r]=r;
#ifndef WINDOWS
    pthread_create(&pid,NULL,(void*)chat_thread,&threads[r]);
#else
    _beginthread((void*)chat_thread,0,&threads[r]);
#endif
  }

  /* chat_thread(); */

  clilen=sizeof(cli_addr);

#ifdef ENABLE_SSL
  int select_sockfd=(sockfd>sockfd_ssl)?sockfd+1:sockfd_ssl+1;
#endif

  while(1)
  {
#ifdef ENABLE_SSL
    // Should probably fix this.  This is really ugly to look at.
    fd_set readset;
    struct timeval tv;
    int t;

    tv.tv_sec=5;
    tv.tv_usec=0;

    FD_ZERO(&readset);
    if (sockfd!=0) { FD_SET(sockfd, &readset); }
    if (sockfd_ssl!=0) { FD_SET(sockfd_ssl, &readset); }

    t=select(select_sockfd, &readset, NULL, NULL, &tv);
    if (t<=0) continue;

    if (FD_ISSET(sockfd, &readset))
    {
      newsockfd=accept(sockfd,(struct sockaddr *)&cli_addr,&clilen);
      if (newsockfd<0)
      {
#ifdef DEBUG
        if (server_info->debug==1) { printf("Server accept error.\n"); }
#endif
        socket_errors++;
        if (socket_errors==5) break;
      }
        else
      {
#ifdef DEBUG
        if (server_info->debug==1)
        {
          printf("New socket accepted\n");
          fflush(stdout);
        }
#endif
        init_user(newsockfd, &cli_addr, clilen, SSL_OFF);
        socket_errors=0;
      }
    }

    if (FD_ISSET(sockfd_ssl, &readset))
    {
      newsockfd=accept(sockfd_ssl,(struct sockaddr *)&cli_addr,&clilen);
      if (newsockfd<0)
      {
#ifdef DEBUG
        if (server_info->debug==1) { printf("Server accept error.\n"); }
#endif
        socket_errors++;
        if (socket_errors==5) break;
      }
        else
      {
#ifdef DEBUG
        if (server_info->debug==1)
        {
          printf("New SSL socket accepted\n");
          fflush(stdout);
        }
#endif
        init_user(newsockfd, &cli_addr, clilen, SSL_ON);
        socket_errors=0;
      }
    }
#else
    // NO SSL

    newsockfd=accept(sockfd,(struct sockaddr *)&cli_addr,&clilen);

    if (newsockfd<0)
    {
#ifdef DEBUG
      if (server_info->debug==1) { printf("Server accept error.\n"); }
#endif
      socket_errors++;
      if (socket_errors==5) break;
    }
      else
    { 
#ifdef DEBUG
      if (server_info->debug==1)
      {
        printf("New socket accepted\n");
        fflush(stdout);
      }
#endif
      init_user(newsockfd, &cli_addr, clilen, SSL_OFF);
      socket_errors=0;
    }
#endif
  }

#ifdef ENABLE_SSL
  SSL_CTX_free(server_info->ssl_ctx);
#endif
  free(server_info);

  return 0;
}

